/***************************************************************************
                          Vector.cc  -  description
                             -------------------
    begin                : Fri May 12 2000
    copyright            : (C) 2000 by Sheldon Lee Wen
    email                : sheldonl@linuxstart.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/* source code for a generic Vector class.  The header is Vector.h
 *
 * A dynamic array implementation of a vector that stores generic (void) 
 * pointers.
 *
 * Written by: Rob Boyd   rboyd@telusplanet.net
 * 
 */

#include "Vector.h"


Vector::Vector(int s = 50) : sz(s)
{
  curSize = numRemoved = 0;
  iter = -1;                // -1 indicates an uninitialized iterator
  data = new void *[sz];

  // initialize the data
  for (int i = 0; i < sz; i++)
    data[i] = NULL;

  assert(data);
}

Vector::Vector(Vector &rhs)
{
   if (&rhs != this) {
      sz = rhs.sz;
      curSize = rhs.curSize;
      numRemoved = rhs.numRemoved;
      iter = rhs.iter;

      // copy the data pointers. Note: This does not copy the data!
      for (int i = 0; i < sz; i++)
         data[i] = rhs.data[i];
   }
}

/*
 * Returns the index of elem in the vector, and -1 if the elem does not occur
 * in the vector.
 */

int Vector::indexOf(void *elem)
{
  int i;

  for(i = 0; i < curSize; i++)
      if (elem == data[i])
	return (i);

  return (-1);
}

/* 
 * Returns 0 if the element is not contained in the vector, 1 if it is.
 */

int Vector::contains(void *elem)
{
  return (indexOf(elem) != -1);
}

/*
 * Will remove all elements from the vector and leave it at it's current size
 */

void Vector::removeAllElements()
{
  for (int i = 0; i < sz; i++)
    data[i] = NULL;

  curSize = numRemoved = 0;
  iter = -1;       // iterator is no longer valid, needs to be reset
}

/*
 * Returns the elements at index, or NULL if there is an error or no element
 * there.
 */

void *Vector::elementAt(int index)
{
  if ((index >= 0) && (index < curSize))  // index is within valid range
    return data[index];

  return NULL;
}

/*
 * Returns the first element in the vector, or NULL if the vector is empty
 *
 * NOTE: The first element does not necessarily have an index of 0
 */

void *Vector::firstElement()
{
  int i;

  for (i = 0; i < curSize; i++)
    if (data[i])         // if it is non-null
      return (data[i]);

  return NULL;           // the vector is empty
}

/*
 * Returns the last element, or NULL if the vector is empty
 */

void *Vector::lastElement()
{
  int i;

  for (i = curSize - 1; i >= 0; i--)
    if (data[i])       
      return (data[i]);

  return NULL;                    // vector is empty
}

/*
 * Will remove the element at index, if the index is out of range, or there
 * is no element at that index, nothing is done.
 */

void Vector::removeElementAt(int index)
{
  if ((index < 0) || (index >= curSize) || (data[index] == NULL))
    return;

  data[index] = NULL;   // remove the item

  numRemoved++;         // increment numRemoved counter

  if (numRemoved > (sz / 4))  // if more than 1/4 of the vector space is wasted
    setSize(sz);              // repack the vector

}

/*
 * Will remove the element 'elem' from the vector if it exists within it.
 */

void Vector::removeElement(void *elem)
{
  int i;

  for (i = 0; i < sz; i++)
  {  if (data[i] == elem)        // found it
     {
        data[i] = NULL;
        numRemoved++;
        if (numRemoved > (sz / 4))  // more than 1/4 wasted space
           setSize(sz);              // need to repack
        curSize--;
     }
  }
}

void Vector::addElement(void *elem)
{
  if (curSize >= sz)           // need more space
    setSize(sz + (sz /2 ));    // increase size by 50%

  data[curSize++] = elem;
}

/*
 * Returns 1 if element at index was successfully set, 0 otherwise (i.e. if
 * index is out of bounds.
 */

int Vector::setElementAt(void *elem, int index)
{
  if ((index < 0) || (index >= sz))
    return (0);

  data[index] = elem;

  return (1);
}

/*
 * Returns 1 if element successfully inserted, 0 otherwise (i.e. index out of
 * bounds.
 */

int Vector::insertElementAt(void *elem, int index)
{
  void **newData;

  if ((index < 0) || (index >= sz))
    return (0);

  if (curSize >= sz)   // we need more space...
    {
      int newSize = sz + (sz / 2);      // increase size by 50%
      newData = new void *[newSize];
      sz = newSize;
    }
  else
    newData = new void *[sz];

  assert(newData != NULL);      // we have enough memory

  // move old data into new space..
  memcpy(newData, data, index * sizeof(void *));

  // insert new element
  newData[index] = elem;

  // copy remainder of oldData
  memcpy(&(newData[index + 1]), &(data[index]),
	                        (sz - index + 1) * sizeof(void *));

  // free up old space, and assign new data into Vector
  delete [] data;

  data = newData;
  curSize++;

  return (1);
}

/*
 * This function will set the capacity of the vector to it's current contents,
 * and "shrink out" removed elements -- essentially reducing the vector to it's
 * minimum size.
 */

void Vector::trimToSize()
{
  setSize(curSize - numRemoved);
}

/*
 * setSize(int) -- sets the size of the vector to the integer passed in.  This
 * function also "compresses" the data in the vector to get rid of fragmented
 * space caused by removal of elements.  The result of all this is that any
 * empty space will be at the end of the vector, curSize and numRemoved will be
 * reset, and the iterator is rendered no longer valid.
 *
 * NOTE: if you setSize to a size smaller than the number of elements in the
 *       vector, the contents of the vector will be truncated so that it only
 *       contains those elements that fit in the new size.
 */

void Vector::setSize(int s)
{
  void **newData;
  int  i = 0;

  newData = new void *[s];
  assert (newData);

  iBegin();         // use the iterator to transfer contents of data to newData

  while (i < s)
    {
      void *tmp;
      tmp = getNext();

      if (!tmp)  // we've hit the end of the vector
	break;

      newData[i++] = tmp;
    }

  delete [] data;
  data = newData;

  // clean up the stats...
  sz = s;
  curSize = i;
  numRemoved = 0;

}

/********* iterator functions ************/

void Vector::iBegin()
{
  iter = 0;
}

void Vector::iEnd()
{
  iter = curSize - 1;  // the last element entered
}

void *Vector::getNext()
{
  if (iter < 0)  // not initialized
    return NULL;

  while ((iter < curSize) && (data[iter] == NULL))
    iter++;

  if (iter < curSize)     // must have the next element
    return (data[iter++]);

  // otherwise we've hit the end of the vector
  return NULL;
}

void *Vector::getPrev()
{
  while ((iter >= 0) && (data[iter] == NULL))
    iter--;

  if (iter >= 0)  // found something
    return (data[iter--]);

  // else we've hit the beginning of the vector
  return NULL;
}
